#include "stdafx.h"
#include "bibobApp.h"
#include "bbstuff.h"

static int gethexdigit(int d)
{
	static char hdigits[] = _T("0123456789ABCDEF");
	char *found = strchr(hdigits, d);
	return (found - hdigits);
}


static bool isFen(u64 &bb, const CString &s, u32 piecemask) {
	static char sFenChars[]  = _T("__KkPpNnBbRrQq12345678");
	static u64 one = 1;
	int rank, file, pc, n, i=0;
	char *fptr;

	n = s.GetLength();

	for (bb = 0, rank = 7; rank>= 0; rank--, i++) {
		for (file = 0; file < 8; file++) {
			if ( i >= n-1 )
				return false;
			fptr = strchr(sFenChars, s.GetAt(i++));
			if ( !fptr )
				return false;
			pc = (int)(fptr - sFenChars);
			if (pc <= 13) {
				if ( piecemask & (1<<pc) )
					bb |= one << ((rank<<3) + file);
			} else {
				file += pc - 14;
			}
		}
		if (rank > 0 && s.GetAt(i) != '/' )
			return false;
	}
	return true;
}

bool isFen(const CString &s) {
	u64 bb;
	return isFen(bb, s, 0);
}

u64 hexdecfenBB(const CString &s, u32 piecemask, int calcmode, int &kind)
{
	u64 bb;
	int n = s.GetLength();
	int i = 0;
	kind = DONT_KNOW;
	bool hex = false;
	if ( n > 0 ) {
		if ( n > 2 && s.GetAt(0) == '0' && toupper(s.GetAt(1)) == 'X') {
			hex = true;
			kind = IS_HEX;
			i = 2;
		} else if ( isFen(bb, s, piecemask) ) {
			kind = IS_FEN;
			return bb;
		} else if ( s.FindOneOf(_T(".")) >= 0 ) {
			union64 dbl;
			dbl.dbl = atof(s);
			bb = dbl.bb;
			kind = IS_DBL;
			return bb;
		} else if ( toupper(s.GetAt(n-1)) == 'H' || s.FindOneOf(_T("abcdefABCDEF")) >= 0) {
			kind = IS_HEX;
			hex = true;
		} else {
			if ( calcmode == DBL ) {
				union64 dbl;
				dbl.dbl = atof(s);
				bb = dbl.bb;
				kind = IS_DBL;
			} else {
				bb = _atoi64(s);
				kind =  IS_DEC;
			}
			return bb;
		}
	}
	for (bb = 0; i < n; i++) {
		int c = gethexdigit(toupper(s.GetAt(i)));
		if ( c < 0 )
			break;
		bb *= 16;
		bb += c;
	}
	return bb;
}

void bb2hexStr(u64 bb, CString &hex) {
	hex.Format("0x%08X%08X", (unsigned int)(bb>>32), (unsigned int)bb);
}

void bb2BoardStr(u64 bb, CString &str) {
	const u64 one = 1;
	str += "\r\n";
	for (int r = 7; r >= 0; r--)
	{
		for (int f = 0; f < 8; f++)
		{
			int sq = 8*r + f;
			u64 sqbb = one << sq;
			if ( bb & sqbb)
			{
				str += " 1";
			}
			else
			{
				str += " .";
			}
		}
		str += "\r\n";
	}

}
	
 	
void bb2decStr(u64 bb, CString &dec) {
	static const char* decformats[2] = {"%s", "-%s"};
	char buff[32];
	int i, j=0;
	int isNeg = 0;
	buff[31] = '\0';
	buff[30] = '0';
	if ( bb & 0x8000000000000000 ) {
		isNeg = 1;
		bb = -(i64)bb;
	}
	for (i=30; i > 0 && bb; i--) {
		j = 1;
		buff[i] = '0' + (char)(bb % 10);
		bb /= 10;
	}
	dec.Format(decformats[isNeg], buff+i+j);
}

// 64*64bit -> upper 64bit of 128-bit product
u64 mulh(u64 bb1, u64 bb2) 
{
	u64 l1 = bb1 & 0xffffffff;
	u64 l2 = bb2 & 0xffffffff;
	u64 h1 = bb1 >> 32;
	u64 h2 = bb2 >> 32;
	u64 l  = (l1 * l2) >> 32;
	u64 m1 = l1 * h2 + l;
	u64 m2 = l2 * h1;
	u64 m  = m1 + m2;
	return h1 * h2 + (m>>32) + (0x100000000 & -(m<m1));
}

// SWAR population count
u32 popCount (u64 bb) {	
	u32 w = (u32)(bb >> 32), v = (u32)bb;
	v = v - ((v >> 1) & 0x55555555);
	w = w - ((w >> 1) & 0x55555555);
	v = (v & 0x33333333) + ((v >> 2) & 0x33333333);
	w = (w & 0x33333333) + ((w >> 2) & 0x33333333);
	v += w;
	v = (v & 0x0F0F0F0F) + ((v >> 4) & 0x0F0F0F0F);
	v = (v * 0x01010101) >> 24;
	return v;
}

u64 add32(u64 a, u64 b) {
	u64 a1 = a & 0x00000000ffffffff;
	u64 b1 = a & 0xffffffff00000000;
	u64 a2 = b & 0x00000000ffffffff;
	u64 b2 = b & 0xffffffff00000000;
	a1 += a2; 
	b1 += b2;
	a1 &= 0x00000000ffffffff;
	b1 &= 0xffffffff00000000;
	return a1|b1;
}

u64 add16(u64 a, u64 b) {
	u64 a1 = a & 0x0000ffff0000ffff;
	u64 b1 = a & 0xffff0000ffff0000;
	u64 a2 = b & 0x0000ffff0000ffff;
	u64 b2 = b & 0xffff0000ffff0000;
	a1 += a2; 
	b1 += b2;
	a1 &= 0x0000ffff0000ffff;
	b1 &= 0xffff0000ffff0000;
	return a1|b1;
}

u64 add8(u64 a, u64 b) {
	u64 a1 = a & 0x00ff00ff00ff00ff;
	u64 b1 = a & 0xff00ff00ff00ff00;
	u64 a2 = b & 0x00ff00ff00ff00ff;
	u64 b2 = b & 0xff00ff00ff00ff00;
	a1 += a2; 
	b1 += b2;
	a1 &= 0x00ff00ff00ff00ff;
	b1 &= 0xff00ff00ff00ff00;
	return a1|b1;
}

u64 sub32(u64 a, u64 b) {
	u64 a1 = a | 0xffffffff00000000;
	u64 b1 = a | 0x00000000ffffffff;
	u64 a2 = b & 0x00000000ffffffff;
	u64 b2 = b & 0xffffffff00000000;
	a1 -= a2; 
	b1 -= b2;
	a1 &= 0x00000000ffffffff;
	b1 &= 0xffffffff00000000;
	return a1|b1;
}

u64 sub16(u64 a, u64 b) {
	u64 a1 = a | 0xffff0000ffff0000;
	u64 b1 = a | 0x0000ffff0000ffff;
	u64 a2 = b & 0x0000ffff0000ffff;
	u64 b2 = b & 0xffff0000ffff0000;
	a1 -= a2; 
	b1 -= b2;
	a1 &= 0x0000ffff0000ffff;
	b1 &= 0xffff0000ffff0000;
	return a1|b1;
}

u64 sub8(u64 a, u64 b) {
	u64 a1 = a | 0xff00ff00ff00ff00;
	u64 b1 = a | 0x00ff00ff00ff00ff;
	u64 a2 = b & 0x00ff00ff00ff00ff;
	u64 b2 = b & 0xff00ff00ff00ff00;
	a1 -= a2; 
	b1 -= b2;
	a1 &= 0x00ff00ff00ff00ff;
	b1 &= 0xff00ff00ff00ff00;
	return a1|b1;
}


/**
 * Flip a bitboard vertically about the centre ranks.
 * Rank 1 is mapped to rank 8 and vice versa.
 * @param x any bitboard
 * @return bitboard x flipped vertically
 */
u64 flip(u64 x) {
   const u64 k1 = u64(0x00FF00FF00FF00FF);
   const u64 k2 = u64(0x0000FFFF0000FFFF);
   x = ((x >>  8) & k1) | ((x & k1) <<  8);
   x = ((x >> 16) & k2) | ((x & k2) << 16);
   x = ( x >> 32)       | ( x       << 32);
   return x;
}

/**
 * Mirror a bitboard horizontally about the center files.
 * File a is mapped to file h and vice versa.
 * @param x any bitboard
 * @return bitboard x mirrored horizontally
 */
u64 mirror (u64 x) {
   const u64 k1 = u64(0x5555555555555555);
   const u64 k2 = u64(0x3333333333333333);
   const u64 k4 = u64(0x0f0f0f0f0f0f0f0f);
   x = ((x >> 1) & k1) | ((x & k1) << 1);
   x = ((x >> 2) & k2) | ((x & k2) << 2);
   x = ((x >> 4) & k4) | ((x & k4) << 4);
   return x;
}

/**
 * Flip a bitboard about the diagonal a1-h8.
 * Square h1 is mapped to a8 and vice versa.
 * @param x any bitboard
 * @return bitboard x flipped about diagonal a1-h8
 */
u64 diaflip(u64 x)
{
   u64 t;
   const u64 k1 = u64(0x5500550055005500);
   const u64 k2 = u64(0x3333000033330000);
   const u64 k4 = u64(0x0f0f0f0f00000000);
   t  = k4 & (x ^ (x << 28));
   x ^=       t ^ (t >> 28) ;
   t  = k2 & (x ^ (x << 14));
   x ^=       t ^ (t >> 14) ;
   t  = k1 & (x ^ (x <<  7));
   x ^=       t ^ (t >>  7) ;
   return x;
}

/**
 * Flip a bitboard about the antidiagonal a8-h1.
 * Square a1 is mapped to h8 and vice versa.
 * @param x any bitboard
 * @return bitboard x flipped about antidiagonal a8-h1
 */
u64 antflip(u64 x) {
   u64 t;
   const u64 k1 = u64(0xaa00aa00aa00aa00);
   const u64 k2 = u64(0xcccc0000cccc0000);
   const u64 k4 = u64(0xf0f0f0f00f0f0f0f);
   t  =       x ^ (x << 36) ;
   x ^= k4 & (t ^ (x >> 36));
   t  = k2 & (x ^ (x << 18));
   x ^=       t ^ (t >> 18) ;
   t  = k1 & (x ^ (x <<  9));
   x ^=       t ^ (t >>  9) ;
   return x;
}